/* $Id: uves_cal_lingain.c,v 1.17 2009/10/29 06:50:32 amodigli Exp $
 *
 * This file is part of the DETMON Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: amodigli $
 * $Date: 2009/10/29 06:50:32 $
 * $Revision: 1.17 $
 * $Name: uves-4_9_1 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------
                             Includes and Defines
 ----------------------------------------------------------------------------*/
#include <stdbool.h>
#include <cpl.h>

#include "irplib_detmon_lg.h"
#include "irplib_detmon.h"

#include "uves_plugin.h"
//#include "uves_chip_type.h"
#include "uves_dfs.h"
#include "uves_msg.h"
#include "uves_error.h"
#include "uves_utils.h"
#include "uves_utils_wrappers.h"

/* Define here the DO.CATG keywords */

#define UVES_LINGAIN_ON_RAW(it) ((it) == 0 ? "SCREEN_FLAT_DETCHAR_RED" : \
                                 (it) == 1 ? "SCREEN_FLAT_DETCHAR_BLUE" : \
                                 "???")

#define UVES_LINGAIN_OFF_RAW(it) ((it) == 0 ? "BIAS_DETCHAR_RED" : \
                                  (it) == 1 ? "BIAS_DETCHAR_BLUE" : "???")

#define UVES_LINGAIN_DET_LIN_INFO(it) ((it) == 0 ? "DET_LIN_INFO_RED" : \
                                       (it) == 1 ? "DET_LIN_INFO_BLUE" : "???")

#define UVES_LINGAIN_GAIN_INFO(it) ((it) == 0 ? "GAIN_INFO_RED" : \
                                    (it) == 1 ? "GAIN_INFO_BLUE" : "???")

#define UVES_LINGAIN_COEFFS_CUBE(it) ((it) == 0 ? "COEFFS_CUBE_RED" : \
                                      (it) == 1 ? "COEFFS_CUBE_BLUE" : "???")

#define UVES_LINGAIN_BP_MAP_NL(it) ((it) == 0 ? "BP_MAP_NL_RED" : \
                                    (it) == 1 ? "BP_MAP_NL_BLUE" : "???")

#define UVES_LINGAIN_AUTOCORR(it) ((it) == 0 ? "AUTOCORR_RED" : \
                                   (it) == 1 ? "AUTOCORR_BLUE" : "???")

#define UVES_LINGAIN_DIFF_FLAT(it) ((it) == 0 ? "DIFF_FLAT_RED" : \
                                    (it) == 1 ? "DIFF_FLAT_BLUE" : "???")


#define RECIPE_NAME "uves_cal_lingain"

/* Copy here instrument specific keywords which need to be in the PAF file */
#define INSTREGEXP   "ESO INS SETUP ID"

#define PAFREGEXP    "^(" REGEXP "|" INSTREGEXP ")$"
#define UVES_TAG_ON  "SCREEN_FLAT_DETCHAR_RED/BLUE"
#define UVES_TAG_OFF "BIAS_DETCHAR_RED/BLUE"

#define OPT FALSE
#define UVES_PIX2PIX CPL_TRUE
#define UVES_EXTS_RED -1
/*----------------------------------------------------------------------------
			     Functions prototypes
 ----------------------------------------------------------------------------*/

cpl_error_code
uves_cal_lingain_fill_parlist_default(cpl_parameterlist * parlist);

UVES_IRPLIB_RECIPE_DEFINE(uves_cal_lingain, UVES_BINARY_VERSION,
		     /* Replace DETMON_BINARY_VERSION with
			corresponding pipeline macro */
                     uves_cal_lingain_fill_parlist_default(recipe->parameters),
		     /* Replace "uves" with PACKAGE_TARNAME
			when moved into UVES */
                     "Lander de Bilbao", PACKAGE_BUGREPORT, "2008",
                     "Linearity/Gain recipe for the optical domain",
                     irplib_detmon_lg_get_description(RECIPE_NAME, "UVES",
                                                      UVES_TAG_ON,
                                                      UVES_TAG_OFF));


/*---------------------------------------------------------------------------*/
/*
  @brief    Interpret the command line options and execute the data processing
  @param    frameset    the frames list
  @param    parlist     the parameters list
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int
uves_cal_lingain(cpl_frameset            * frameset,
                 const cpl_parameterlist * parlist)
{


   int red;
   bool found_bias = false;
   bool found_flat = false;
   cpl_error_code error=0;
   int processed=0;
   cpl_parameter* p=NULL;
   cpl_parameterlist* plist=NULL;

   check_nomsg(plist=uves_parameterlist_duplicate(parlist));

   for (red = 0; red <= 1; red++){

      check_nomsg(found_bias = cpl_frameset_find(frameset,UVES_LINGAIN_ON_RAW(red)));
      check_nomsg(found_flat = cpl_frameset_find(frameset,UVES_LINGAIN_OFF_RAW(red)));

      if((found_bias) && (found_flat)) {

         if(red == 0) {
            uves_msg("Update exts param for RED ARM data");

            check_nomsg(p=cpl_parameterlist_find(plist,PACKAGE "." RECIPE_NAME ".exts"));

            cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);
            check_nomsg(error = cpl_parameter_set_int(p, UVES_EXTS_RED));
            cpl_ensure_code(!error, error);




         }
         processed++;

         cpl_propertylist * lintbl =
            irplib_detmon_fill_prolist(UVES_LINGAIN_DET_LIN_INFO(red), "TYPE", "TECH", CPL_TRUE);

         cpl_propertylist * gaintbl =
            irplib_detmon_fill_prolist(UVES_LINGAIN_GAIN_INFO(red),    "TYPE", "TECH", CPL_TRUE);

         cpl_propertylist * coeffscube =
            irplib_detmon_fill_prolist(UVES_LINGAIN_COEFFS_CUBE(red),  "TYPE", "TECH", CPL_TRUE);

         cpl_propertylist * bpm =
            irplib_detmon_fill_prolist(UVES_LINGAIN_BP_MAP_NL(red),    "TYPE", "TECH", CPL_TRUE);

         cpl_propertylist * corr =
            irplib_detmon_fill_prolist(UVES_LINGAIN_AUTOCORR(red),     "TYPE", "TECH", CPL_TRUE);

         cpl_propertylist * diff_flat =
            irplib_detmon_fill_prolist(UVES_LINGAIN_DIFF_FLAT(red),    "TYPE", "TECH", CPL_TRUE);

         error = irplib_detmon_lg(frameset,
                                              plist,
                                              UVES_LINGAIN_ON_RAW(red),
                                              UVES_LINGAIN_OFF_RAW(red),
                                              RECIPE_NAME,
                                              "uves",
                                              PAFREGEXP,
                                              lintbl,
                                              gaintbl,
                                              coeffscube,
                                              bpm,
                                              corr,
                                              diff_flat,
                                              PACKAGE "/" PACKAGE_VERSION,
                                              NULL, NULL, OPT);

         cpl_propertylist_delete(lintbl);
         cpl_propertylist_delete(gaintbl);
         cpl_propertylist_delete(coeffscube);
         cpl_propertylist_delete(bpm);
         cpl_propertylist_delete(corr);
         cpl_propertylist_delete(diff_flat);
         lintbl = 0;
         gaintbl = 0;
         coeffscube = 0;
         bpm = 0;
         corr = 0;
         diff_flat = 0;
         check_nomsg(error);
      }
   }
   if(processed>0) {
      /* Propagate the error, if any */
      cpl_ensure_code(!error, error);
   } else {
      uves_msg("No data has been processed.");
      uves_msg("Pls check if input frame set contains valid tags");
   }



  cleanup:
   //uves_free_parameterlist(&parlist);
   return error;
}

cpl_error_code
uves_cal_lingain_fill_parlist_default(cpl_parameterlist * parlist)
{

    cpl_error_code error =
        irplib_detmon_lg_fill_parlist_opt_default(parlist, RECIPE_NAME, PACKAGE);
    cpl_parameter * p;

    cpl_ensure_code(!error, error);

    /* FIXME: Put the parameter names and their new default values in two arrays
       and iterate through them */
    p = cpl_parameterlist_find(parlist, PACKAGE "." RECIPE_NAME ".pix2pix");

    cpl_ensure_code(p != NULL, CPL_ERROR_DATA_NOT_FOUND);

    error = cpl_parameter_set_default_bool(p, UVES_PIX2PIX);


    cpl_ensure_code(!error, error);

    return CPL_ERROR_NONE;
}


