# import the needed modules
try:
  import numpy
  import pyfits
  import wx
  import matplotlib
  matplotlib.interactive(False)
  matplotlib.use('WXAgg')
  from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
  from matplotlib.backends.backend_wx import NavigationToolbar2Wx
  from matplotlib.figure import Figure
  from matplotlib.pyplot import gcf, setp
  import_sucess = 'true'

#NOTE for developers: 
# -If you want to modify the current script to cope
#  with different parameters, this is the function to modify:
#  setInteractiveParameters()
# -If you want to modify the current script to read different data from
#  the input FITS, this is the function to modify:
#  readFitsData()                  (from class DataPlotterManager) 
# -If you want to modify the current script to modify the plots (using the same
#  data),  this is the function to modify:
#  plotProductsGraphics()          (from class DataPlotterManager)
# -If you want to modify the text that appears in the "Help" button,
#  this is the function to modify:
#  setWindowHelp()
# -If you want to modify the tittle of the window, modify this function:
#  setWindowTittle()


  # This function specifies which are the parameters that should be presented
  # in the window to be edited.
  # Note that the parameter has to be also in the in_sop port (otherwise it 
  # won't appear in the window) 
  # The descriptions are used to show a tooltip. They should match one to one
  # with the parameter list 
  # Note also that parameters have to be prefixed by the 'recipe name:'
  def setInteractiveParameters():
    param_list = ['uves_cal_orderpos:use_guess_tab',
                 'uves_cal_orderpos:radx',
                 'uves_cal_orderpos:rady',
                 'uves_cal_orderpos:mmethod',
                 'uves_cal_orderpos:backsubgrid',
                 'uves_cal_orderpos:backsubradiusy',
                 'uves_cal_orderpos:backsubkappa',
                 'uves_cal_orderpos:backsubdegx',
                 'uves_cal_orderpos:backsubdegy',
                 'uves_cal_orderpos:samplewidth',
                 'uves_cal_orderpos:minslope',
                 'uves_cal_orderpos:maxslope',
                 'uves_cal_orderpos:sloperes',
                 'uves_cal_orderpos:norders',
                 'uves_cal_orderpos:pthres',
                 'uves_cal_orderpos:tracestep',
                 'uves_cal_orderpos:minthresh',
                 'uves_cal_orderpos:maxgap',
                 'uves_cal_orderpos:maxrms',
                 'uves_cal_orderpos:defpol1',
                 'uves_cal_orderpos:defpol2',
                 'uves_cal_orderpos:kappa']
    param_desc = ['Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc',
                 'Desc']
    return param_list, param_desc

  def setWindowHelp():
    help_text = 'Placeholder for Window specific help'
    return help_text

  def setWindowTittle():
    tittle = 'Uves Interactive Orderpos'
    return tittle

  #This class deals with the reading of data and final plotting.
  class DataPlotterManager:
    # This function will read all the columns, images and whatever is needed
    # from the products. The variables , self.plot_x, self.plot_y, etc...
    # are used later in function plotProductsGraphics().
    # Add/delete these variables as you need (only that plotProductsGraphics()
    # has to use the same names).
    # You can also create some additional variables (like statistics) after
    # reading the files.
    # If you use a control variable (self.xxx_found), you can modify 
    # later on the layout of the plotting window based on the presence of 
    # given input files. 
    # sof contains all the set of frames
    def readFitsData(self, sof):
      #Control variable to check if the interesting files where at the input
      self.order_table_blue_found = False
      #Read all the products
      for frame in sof:
        if frame == '' :
          continue
        hdulist = pyfits.open(frame)
        pro_catg = hdulist[0].header['ESO PRO CATG']
        if pro_catg == "ORDER_TABLE_BLUE" :
          self.order_table_blue_found = True
          self.plot_x        = hdulist[1].data.field('X')
          self.plot_y        = hdulist[1].data.field('Y')
          self.plot_residual = hdulist[1].data.field('Residual')
          self.plot_order    = hdulist[1].data.field('Order')
          self.npred         = hdulist[0].header['ESO QC ORD NPRED']
          self.ndet          = hdulist[0].header['ESO QC ORD NDET']
          self.det_nx        = hdulist[0].header['ESO DET OUT1 NX']
          self.det_ny        = hdulist[0].header['ESO DET OUT1 NY']
      #Some statistics about the values          
      #We compute them only if the input data was read
      if self.order_table_blue_found == True :
        self.filt_residual   = self.plot_residual[numpy.isfinite(self.plot_residual)]
        self.mean_residual   = numpy.mean(self.filt_residual)
        self.median_residual = numpy.median(self.filt_residual)
        self.std_residual    = numpy.std(self.filt_residual)
        self.min_x        = numpy.nanmin(self.plot_x)
        self.max_x        = numpy.nanmax(self.plot_x)
        self.min_y        = numpy.nanmin(self.plot_y)
        self.max_y        = numpy.nanmax(self.plot_y)


    # This function creates all the subplots. It is responsible for the plotting 
    # layouts. 
    # There can different layouts, depending on the availability of data
    # Note that subplot(I,J,K) means the Kth plot in a IxJ grid 
    # Note also that the last one is actually a box with text, no graphs.
    def addSubplots(self, figure):
      if self.order_table_blue_found == True :
        self.subplot1 = figure.add_subplot(2,2,1)
        self.subplot2 = figure.add_subplot(2,2,2)
        self.subplot3 = figure.add_subplot(2,2,3)
        self.subplot4 = figure.add_subplot(2,2,4)
      else : 
        self.subtext_nodata = figure.add_subplot(1,1,1)
          
    # This is the function that makes the plots.
    # Add new plots or delete them using the given scheme.
    # The data has been already stored in self.plot_x, self.plot_xdif, etc ...
    # It is mandatory to add a tooltip variable to each subplot.
    # One might be tempted to merge addSubplots() and plotProductsGraphics().
    # There is a reason not to do it: addSubplots() is called only once at
    # startup, while plotProductsGraphics() is called always there is a resize.
    def plotProductsGraphics(self, figure, canvas):
      if self.line_guess_tab_blue_found == True :
        #First plot
        self.subplot1.scatter(self.plot_x, self.plot_residual, s=0.5, color='blue')
        self.subplot1.set_xlabel('X [pix]')
        self.subplot1.set_ylabel('(Y - YFIT) [pix]')
        self.subplot1.set_xlim(self.min_x, self.max_x)
        self.subplot1.grid(True)
        self.title1 = 'Y - YFIT: mean=%#.3g med=%#.3g rms=%#.3g'% (self.mean_residual,self.median_residual,self.std_residual) 
        self.subplot1.set_title(self.title1, fontsize=12)
        self.subplot1.tooltip='Residual vs X'

        #Second plot
        self.subplot2.scatter(self.plot_y, self.plot_residual, s=0.5, color='blue')
        self.subplot2.set_xlabel('Y [pix]')
        self.subplot2.set_ylabel('(Y - YFIT) [pix]')
        self.subplot2.set_xlim(self.min_y, self.max_y)
        self.subplot2.grid(True)
        self.title2 = 'Y - YFIT: mean=%#.3g med=%#.3g rms=%#.3g'% (self.mean_residual,self.median_residual,self.std_residual) 
        self.subplot2.set_title(self.title2, fontsize=12)
        self.subplot2.tooltip='Residual vs Y'
      
        #Third plot
        self.subplot3.scatter(self.plot_order, self.plot_residual, s=0.5, color='blue')
        self.subplot3.set_xlabel('ORDER')
        self.subplot3.set_ylabel('(Y - YFIT) [pix]')
        self.subplot4.set_xlim(0., self.ndet + 1.)
        self.subplot3.grid(True)
        self.title3 = 'Y - YFIT: mean=%#.3g med=%3.3g rms=%#.3g'% (self.mean_residual,self.median_residual,self.std_residual) 
        self.subplot3.set_title(self.title3, fontsize=12)
        self.subplot3.tooltip='Residual vs Order'
      
        #Fourth plot
        self.subplot4.scatter(self.plot_x, self.plot_y, s=0.5, color='blue')
        self.subplot4.set_xlabel('X [pix]')
        self.subplot4.set_ylabel('Y [pix]')
        self.subplot4.set_xlim(0., self.det_ny)
        self.subplot4.set_ylim(0., self.det_nx)
        self.subplot4.grid(True)
        self.title4 = 'Order Traces' 
        self.subplot4.set_title(self.title4, fontsize=12)
        self.subplot4.tooltip='Y vs X'
  
      else :
        #Data not found info
        self.subtext_nodata.set_axis_off()
        self.text_nodata = 'Order table  not found in the products (PRO.CATG=ORDER_TABLE_BLUE)' 
        self.subtext_nodata.text(0.1, 0.6, self.text_nodata, color='#11557c', fontsize=18,
                           ha='left', va='center', alpha=1.0)
        self.subtext_nodata.tooltip='Order table  not found in the products'
  
  
  #NOTE for developers: If you just want to use this script as a template,
  #do not modify anything below
  

  #This class is responsible for the plotting window (which uses matplotlib)
  class PipelinePlotsWindow(wx.Window):
    def __init__(self, frame, sof):
      wx.Window.__init__(self, frame)
      
      #Read the data
      self.dataPlotManager = DataPlotterManager()
      self.dataPlotManager.readFitsData(sof)
      
      #Get the matplotlib canvas
      self.figure = Figure()
      self.canvas = FigureCanvasWxAgg(self, -1, self.figure)
      self.subplotsCreated = False
      self.draw()
      self.canvas.SetToolTip(wx.ToolTip(tip=''))
      self.toolbar = NavigationToolbar2Wx(self.canvas)
      frame.SetToolBar(self.toolbar)
      self.canvas.mpl_connect('axes_enter_event',self.onEnterAxes)
      self.canvas.mpl_connect('axes_leave_event',self.onLeaveAxes)
  
      #React to resizing events 
      self.Bind(wx.EVT_SIZE, self.sizeHandler)
  
    def onEnterAxes(self, event):
      self.canvas.GetToolTip().SetTip(event.inaxes.tooltip)
      self.canvas.GetToolTip().Enable(True)

    def onLeaveAxes(self, event):
      self.canvas.GetToolTip().Enable(False)
  
    def sizeHandler(self, *args, **kwargs):
      self.canvas.SetSize(self.GetSize())
  
    def draw(self):
      if self.subplotsCreated == False:
        self.dataPlotManager.addSubplots(self.figure)
        self.subplotsCreated = True
        
      self.dataPlotManager.plotProductsGraphics(self.figure, self.canvas)
      self.figure.subplots_adjust(wspace=0.40, hspace=0.55, top=0.95, bottom=0.05, left=0.1, right=0.95)

    def repaint(self):
      self.canvas.draw()
  

  #This class is used to display the label and edit box of a parameter
  class ParamWidget:
    def __init__(self, parent, label, paramValue, description):
      self.paramLabel = wx.StaticText(parent, label=label)
      self.paramEditText = wx.TextCtrl(parent, -1, style=wx.TE_PROCESS_ENTER)
      self.paramEditText.SetValue(paramValue)
      self.paramEditText.SetToolTipString(description)
  
      #This binding is unnecessary, since the OnRepeat directly access the
      #GetValue() of paramEditText. I keep it here in case that
      #the desired behaviour is that you have to press Enter to actually edit the
      #value
      self.paramEditText.Bind(wx.EVT_TEXT_ENTER, self.paramEditTextHandler)
  
      self.paramValue = paramValue
  
    def paramEditTextHandler(self, evt):
      value = self.paramEditText.GetValue()
      self.paramValue = value
  
  #This is the window (frame in WX jargon) used to plot everything 
  class PipelineInteractiveFrame(wx.Frame):
  
    def __init__(self, inputs, outputs):
      wx.Frame.__init__(self, parent=None, title=setWindowTittle(), 
                        size=(1324, 900))
  
      #Initialize the inputs and outputs
      self.inputs  = inputs
      self.outputs = outputs
  
      #Get all the frames
      in_sof_splitted = self.inputs.in_sof.split(',')
      self.sof = list()
      for frame in in_sof_splitted:
        self.sof.append(frame[0:frame.rfind(':')])
        
      #Set the parameters to be plotted
      self.paramList, self.paramDesc = setInteractiveParameters()
      
      #Split the in_sop port
      self.input_sop = inputs.in_sop.split(',')
  
      #The plotting window
      self.pipePlotsWindow = PipelinePlotsWindow(self, self.sof)
  
      #The control area
      self.controlArea = wx.BoxSizer(wx.HORIZONTAL)
  
      #The area for the buttons
      self.buttonArea = wx.BoxSizer(wx.VERTICAL)
  
      #The area for the parameters
      self.parameterArea = wx.StaticBoxSizer(wx.StaticBox(self, -1,'Parameters'),
                                             wx.VERTICAL)
      self.parameterGrid = wx.GridBagSizer(4,4)
  
      #The widgets for the parameters
      self.paramNames   = list()
      self.paramValues  = list()
      self.paramWidgets = list()
      iParam = 0
      for param, desc in zip(self.paramList, self.paramDesc):
        for param_in_sop in self.input_sop:
          if param_in_sop.partition('=')[0] == param:
            self.paramNames.append(param_in_sop.partition('=')[0] )
            self.paramValues.append(param_in_sop.partition('=')[2]) 
            paramName = param_in_sop.partition('=')[0].partition(':')[2]
            paramWidget = ParamWidget(self, label=paramName, \
                                     paramValue=self.paramValues[iParam],
                                     description=desc)
            self.paramWidgets.append(paramWidget)
            iParam = iParam + 1
            
      # The buttons
      self.contButton = wx.Button(self, label="Continue wkf")
      self.Bind(wx.EVT_BUTTON, self.OnCont, self.contButton)
      self.repeatButton = wx.Button(self, label="Re-run recipe")
      self.Bind(wx.EVT_BUTTON, self.OnRepeat, self.repeatButton)
      self.exportpdfButton = wx.Button(self, label="Print plot")
      self.Bind(wx.EVT_BUTTON, self.OnPrint, self.exportpdfButton)
      self.helpButton = wx.Button(self, label="Help")
      self.Bind(wx.EVT_BUTTON, self.OnHelp, self.helpButton)
  
      #Pack parameter widgets
      iParam = 0
      for paramWidget in self.paramWidgets:
        self.parameterGrid.Add(paramWidget.paramLabel, (iParam, 0),
            flag=wx.TOP | wx.LEFT | wx.BOTTOM, border=5)
        self.parameterGrid.Add(paramWidget.paramEditText, (iParam, 1),
            flag=wx.EXPAND | wx.LEFT | wx.RIGHT, border=5)
        iParam = iParam + 1
  
      #Pack buttons
      self.buttonArea.Add(self.contButton, 0, \
          wx.ALIGN_CENTER | wx.ALL, border=5)
      self.buttonArea.Add(self.repeatButton, 0, \
          wx.ALIGN_CENTER | wx.ALL, border=5)
      self.buttonArea.Add(self.exportpdfButton, 0, \
          wx.ALIGN_CENTER | wx.ALL, border=5)
      self.buttonArea.Add(self.helpButton, 0, \
          wx.ALIGN_CENTER | wx.ALL, border=5)
  
      #Pack the parameter grid
      self.parameterArea.Add(self.parameterGrid, 0, \
            flag=wx.EXPAND | wx.ALIGN_CENTER | wx.ALL, border=5) 
  
      #Pack the parameters area
      self.controlArea.Add(self.parameterArea, 0, \
          wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_LEFT | wx.ALL, border=5)
  
      #Pack the button area
      self.controlArea.Add(self.buttonArea, 0, \
          wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT | wx.ALL, border=5)
  
      #The Sizer which packs everything
      sizer = wx.BoxSizer(wx.HORIZONTAL)
      sizer.Add(self.pipePlotsWindow, 1, wx.EXPAND)
      sizer.Add(self.controlArea, 0, wx.EXPAND | wx.ALL, border=5)
      
      self.SetSizer(sizer)
  
    def OnCont(self, event):
      self.outputs.iteration_complete = 'true'
      self.outputs.out_sof            = self.inputs.in_sof
      self.outputs.out_sop            = self.inputs.in_sop
  
      self.Close(True)

    def OnRepeat(self, event):
      #This will update the output parameters based on the entry boxes.
      outputs.out_sop_loop = ''
      for param_in_sop in self.input_sop:
        finalValue = param_in_sop
        for i in range(len(self.paramWidgets)):
          if param_in_sop.partition('=')[0] == self.paramNames[i]:
              finalValue = self.paramNames[i] + '=' +  \
                           self.paramWidgets[i].paramEditText.GetValue()
        if outputs.out_sop_loop != '':
          self.outputs.out_sop_loop += ','
        self.outputs.out_sop_loop += finalValue
    
      self.outputs.iteration_complete = 'false'
      self.outputs.out_sof_loop       = self.inputs.in_sof_rec_orig
      self.Close(True)

    def OnPrint(self, event):
      self.pipePlotsWindow.canvas.Printer_Print(event=event)
      return

    def OnHelp(self, event):
      window_help_msg = setWindowHelp()
      general_help_msg = """
      
The window has several parts:\n
   1. Plot area. This area shows how good the reduction was performed
   2. Parameter area. This area shows the parameters used for the execution of the recipe. The parameters can be changed if the recipe has to be re-run with new parameters
   3. Button area. These buttons control the interactivity of the window\n
        a) Continue wkf. This button sends the current results to the next recipe in the workflow.\n
        b) Re-run recipe. This button will execute the predict recipe again with the new parameters\n
        c) Print plot. This button will print the plot area\n
        d) Help. This button shows this help
   4. Top toolbar area. These buttons allow certain interactivity with the plots (zoom, shift, layout) as well as exporting to png 
        """
      dlg = wx.MessageDialog(self, window_help_msg + general_help_msg, setWindowTittle()+" Help",
                             wx.OK | wx.ICON_INFORMATION)
      dlg.ShowModal()
      dlg.Destroy()

  #This class presents the wx application interface
  class PipelineInteractiveApp(wx.App):

    def __init__(self, inputs, outputs):
      self.inputs = inputs
      self.outputs = outputs
      wx.App.__init__(self,0)
  
    def OnInit(self):
      'Create the main window and insert the custom frame'
      self.frame = PipelineInteractiveFrame(self.inputs, self.outputs)
      self.frame.Show(True)
      return True

except ImportError:
  import_sucess = 'false'
  print "Error importing modules pyfits, wx, matplotlib"

#This is the 'main' function
if __name__ == '__main__':

  # import reflex modules
  from reflex import *

  # create an option parser
  parser = ReflexIOParser()

  # define inputs (Note: you must define at least long option)
  parser.add_option("-i", "--in_sof", dest="in_sof")
  parser.add_option("-p", "--in_sop", dest="in_sop")
  parser.add_option("-g", "--in_sof_rec_orig", dest="in_sof_rec_orig")
  parser.add_option("-e", "--enable", dest="enable")

  # define outputs (Note: you must define at least long option)
  parser.add_output("-o", "--out_sof", dest="out_sof")
  parser.add_output("-q", "--out_sop", dest="out_sop")
  parser.add_output("-f", "--out_sof_loop", dest="out_sof_loop")
  parser.add_output("-l", "--out_sop_loop", dest="out_sop_loop")
  parser.add_output("-c", "--iteration_complete", dest="iteration_complete")

  # get inputs from the command line
  (inputs, args) = parser.parse_args()
  # get output variables
  outputs = parser.get_outputs()

  #Input validation
  if inputs.enable != 'true' and inputs.enable != 'false' :
    raise ValueError('Invalid --enable option. Use true or false')  

  #Check if import failed or not
  if import_sucess == 'false' :
    inputs.enable = 'false'

  #Open the interactive window
  if inputs.enable == 'true' :
    app = PipelineInteractiveApp(inputs, outputs)
    app.MainLoop()
  else :
    outputs.iteration_complete = 'true'
    outputs.out_sof            = inputs.in_sof
    outputs.out_sop            = inputs.in_sop

  # print outputs
  parser.print_outputs()

  sys.exit()


